/* Computer the pairwise scores between SNPs and mRNA or mir expression data.
   Difference from compute_pairwise_scores_cluster_snps2.c: consider the inter-gene pairs.
   Import the intra-gene SNP-mRNA expression pairs.
   Only consider the SNPs that appear in the intra-gene SNP-mRNA expression pairs.
   Build associations with these SNPs and mRNA or microRNA expression data.
   Exclude the intra-gene SNP associations that appeared.
   Difference from compute_pairwise_scores_cluster_snps3.c: (1)perform pairwise associations of all SNPs and mRNA expressions, (2)use SNPs as the basic subunits of computations, (3)only load the SNP data within the range designated by the input.
   Difference from compute_pairwise_scores_cluster_snps4.c: correct the log likelihood evaluation when the null model is independent.  This will reduce the ostensibly inflated log likelihood values of poorly associated pairs.
*/



# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "motif.h"
# include "models.h"
# define MAXLINEWIDTH 1000000
# define NAMELENGTH 30
# define LONGNAMELENGTH 100
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

extern int pair_cmp(struct pair *p1, struct pair *p2);
extern char * getitemval2(char *s, int ind, int *length);
extern char *i2a(long num, int *length);
extern double atof2(char *s, int length);
extern char * getitemval3(char *s, int ind, int *length, char sepch);
extern double * ml_pairwise_model(int nsamples, double *x, double *y, int mode);
extern double * ml_pairwise_model2(int nsamples, double *x, double *y, int mode);
extern double new_missing_corrcoef(int nsamples, double *x, double *y);
extern double * load_matrix(char *filename, int *nrowspnt, int *ncolspnt);
extern struct matdata * load_highRes_data(char *filename, int *nprobespnt, int *nsamplespnt);
extern struct matdata * load_highRes_data_range(char *filename, int startindex, int endindex, int *nprobespnt, int *nsamplespnt);
extern double * ml_pairwise_model_corrected(int nsamples, double *x, double *y, int mode, int nstates);

struct geneseginfo {
  char *name;
  char *snpname;
  int  chr;
  char *dir;
  long startcoord;
  long endcoord;
};

int nintervals=100;
int nmaxiterations=1000, nperm=10000;
double NAval=-1e6, epsilon=0.001, infval=1e20;
double high_thre=0.6, low_thre=0.4;
int ndigits=10;

int main(int argc, char *argv[])
{
  // *(argv+1) -- mode.
  // *(argv+2) -- SNP data.
  // *(argv+3) -- mRNA/mir expression data.
  // *(argv+4) -- correlation coefficient threshold.
  // *(argv+5) -- logratio threshold.
  // *(argv+6) -- pvalue threshold.    
  // *(argv+7) -- increment size of the range number.
  // *(argv+8) -- the number specifying the starting input index.
  // *(argv+9) -- output association scores.
 
  int i, j, k, l, m, n, mode, ngenes, nsamples, length, ninputs, npairs, wrote=0, ncols;
  double val, val2, logratio, pval, *outputdata, *inputdata;
  char *s, *item, ch, *samples, *prefix, *filename, *str, *str2;
  double corrthre, logratiothre, pvalthre, *scores, *retvec;
  long incsize, startind, startindex, inputindex, expindex, endindex, index, geneind, distthre;
  FILE *fp, *fout;
  struct matdata *template;
  struct genemirseginfo *inputs, *genes;
  time_t *tp;

  // Import the mode of expression data.
  // 0: mRNA data.
  // 1: mir data.
  mode=atoi(*(argv+1)); s=malloc(sizeof(char)*MAXLINEWIDTH); 


  // Load the increment size and start index.
  incsize=atoi(*(argv+7)); startind=atoi(*(argv+8));
  startindex=startind*incsize;
  endindex=startindex+incsize-1;
  

  // Import the subset of SNP data.
  // Only the SNPs between [startindex,endindex] are included.
  template=load_highRes_data_range(*(argv+2),startindex,endindex,&ninputs,&nsamples);
  inputs=template->probes;
  samples=template->samples;
  inputdata=template->data;


  // Convert the SNP data into trinary.
  // -1: homozygous of minor alleles.
  // 0: heterozygous.
  // 1: homozygous of major alleles.
  // Restrict them in [0,1].
  for (m=0; m<=(ninputs-1); m++)
    {
      for (n=0; n<=(nsamples-1); n++)
	{
	  val=*(inputdata+m*nsamples+n);
	  if (val>NAval)
	    {
	      if (val==0)
		*(inputdata+m*nsamples+n)=0;
	      else if (val==2)
		*(inputdata+m*nsamples+n)=1;
	      else
		*(inputdata+m*nsamples+n)=0.5;		  
	    }	  
	}
    }

  // Debug
  printf("ninputs=%d\n",ninputs);

  
  // Load mRNA/mir expression data.
  // Assume the genes and samples have the same order as input.  
  template=load_highRes_data(*(argv+3),&ngenes,&nsamples);
  genes=template->probes;
  outputdata=template->data;

  // Debug
  printf("ngenes=%d\n",ngenes);  

  // Load threshold values of logratio and pvalue.
  corrthre=atof(*(argv+4)); logratiothre=atof(*(argv+5)); pvalthre=atof(*(argv+6));

      
  // Generate the file name of the output file.
  // The file name is prefix + number.
  // To facilitate file concatenation, force all numbers to have the same number of digits.
  // Fill 0s between the start and the index number.
  prefix=malloc(sizeof(char)*MAXLINEWIDTH);
  strncpy(prefix,*(argv+9),strlen(*(argv+9))); *(prefix+strlen(*(argv+9)))='\0';
  val=(double)startind;
  if (startind==0)
    k=1;
  else
    {
      val=log10(val); val=floor(val)+1; k=(int)val;
    }
  l=ndigits-k;
  filename=malloc(sizeof(char)*MAXLINEWIDTH);
  str=malloc(sizeof(char)*MAXLINEWIDTH);
  for (i=0; i<=(l-1); i++)
    *(str+i)='0';
  *(str+l)='\0';
  str2=malloc(sizeof(char)*MAXLINEWIDTH);
  sprintf(str2,"%d\0",startind); m=0;
  strncpy(filename,prefix,strlen(prefix)); *(filename+m+strlen(prefix))='\0'; m+=strlen(prefix);
  strncat(filename,str,strlen(str)); *(filename+m+strlen(str))='\0'; m+=strlen(str);
  strncat(filename,str2,strlen(str2)); *(filename+m+strlen(str2))='\0'; m+=strlen(str2);
        
  if ((fout=fopen(filename,"w"))==NULL)
    {
      printf("Cannot open %s.\n",filename);
      return -1;
    }
  
  // Calculate the pairwise association scores.
  // Get the pairs specified in the range [startindex,endindex].

  tp=malloc(sizeof(time_t)); srand((int)(time(tp)));
  
  // For each SNP m, perform associations with all mRNA expressions.
  for (m=0; m<=(ninputs-1); m++)
    {
      for (n=0; n<=(ngenes-1); n++)
	{
	  int flag=1, funcmode=0, funcsign=0;

	  // Discard if the number of samples with valid SNP and expression entries >= 1/2 of the sample size.
	  k=0;
	  for (i=0; i<=(nsamples-1); i++)	    
	    if ((*(inputdata+m*nsamples+i)>=0)&&(*(outputdata+n*nsamples+i)>=0))
	      k++;
	  if (k<(nsamples/2))
	    flag=0;
	  	  
	  // Discard when the correlation coefficient does not pass the threshold.
	  if (flag==1)
	    {
	      val=new_missing_corrcoef(nsamples,inputdata+m*nsamples,outputdata+n*nsamples);
	      val2=fabs(val);
	      if (val2<corrthre)
		flag=0;
	      else
		{
		  if (val>0)
		    {
		      funcmode=1; funcsign=1;
		    }
		  else
		    {
		      funcmode=2; funcsign=-1;
		    }
		}
	    }

	  // Evaluate the llratio and pvalue.
	  if (flag==1)
	    {
	      retvec=ml_pairwise_model_corrected(nsamples,inputdata+m*nsamples,outputdata+n*nsamples,funcmode,3);
	      
	      if ((*(retvec+0)>=logratiothre)&&(*(retvec+1)<=pvalthre))
		{
		  fprintf(fout,"%d%c%d%c%d%c%f%c%.4e%c%f\n",startindex+m+1,0x09,n+1,0x09,funcmode,0x09,*(retvec+0),0x09,*(retvec+1),0x09,val);
		  wrote=1;
		  
		}
	      free(retvec);
	    }
	}
    }
	  
  fclose(fout);

  // If nothing is written on fout, then delete the file.
  if (wrote==0)
    {
      char *commandstr;
      commandstr=malloc(sizeof(char)*MAXLINEWIDTH);
      sprintf(commandstr,"rm %s\0",filename);
      system(commandstr);
      free(commandstr);
    }
  
  return 0;
}
