/* Generate cliques and clusters from the correlation coefficient matrix of data.
   The cliques are more conservative clusters than the results from hierarchical clustering or spectral clustering.
   Goal: find the explanatory causes for each clique.
   Difference from generate_cliques.c: Sort the correlation coefficients.  Start with the highest threshold and find cliques.  Lower the threshold and incorporate genes to the old cliques or generate new cliques.  Stop when the threshold reaches the limit specified by the input.
   Difference from generate_cliques2.c: exclude the members in the large cluster.
   Difference from generate_cliques3.c: incrementally merge cliques to form clusters, instead of retrieving the connected components.
   In TCGA data, discard the leukemia specific clusters.
   Difference from generate_cliques_clusters.c: (1)do not separate the inputs of gene names and sample names, (2)do not consider the large cluster.
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "motif.h"
# define MAXLINEWIDTH 1000000
# define NAMELENGTH 30
# define LONGNAMELENGTH 100
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

struct intset {
  int nmembers;  // number of members.
  int *members;  // members.
};

extern int pair_cmp(struct pair *p1, struct pair *p2);
extern char * getitemval2(char *s, int ind, int *length);
extern char *i2a(long num, int *length);
extern double atof2(char *s, int length);
extern char * getitemval3(char *s, int ind, int *length, char sepch);
extern double corrcoef(int nsamples, double *x, double *y);
extern char * getnextitemval2(char *s, int *curpos, int *length);
extern double mean(int nsamples, double *x);
extern int triplet_cmp(struct triplet *p1, struct triplet *p2);
extern int inlist(int Nv, int *v, int target);
extern int * find_undir_conn_comps(int nnodes, int *conn);
extern void recursive_labeling(int nnodes, int *conn, int *labels, int targetnode, int targetlabel);

int nintervals=100;
int nmaxiterations=1000, nperm=10000;
double NAval=-1e6, epsilon=0.001, infval=1e20;
double high_thre=0.6, low_thre=0.4;

int main(int argc, char *argv[])
{
  // *(argv+1) -- input data.
  // *(argv+2) -- correlation coefficient threshold lower bound.
  // *(argv+3) -- output cliques.
  // *(argv+4) -- threshold on difference between intra-clique and inter-clique correlation coefficients.
  // *(argv+5) -- output clusters.
  
  int i, j, k, l, m, n, nvars, nsamples, ncliques, length, *clids, curind, flag, flag2, *newclids;
  long *chrs, *coords, *indices, ntriplets, incsize;
  int nclus, *cluids, *G, *cliqueids, nsubset, *subset, *isvalid, *large;
  char *s, *item, *names, *samples, *dirs, *genestrs, ch;
  double *data, val, *corrs, threlwb, corrthre, diffthre, *meanprofile;
  struct pair *pairs;
  struct intset *cliques;
  struct triplet *triplets;
  FILE *fp, *fout;


  // Open the input data file.
  if ((fp=fopen(*(argv+1),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+1));
      return -1;
    }
  
  s=malloc(sizeof(char)*MAXLINEWIDTH); 

  // Load sample info.
  fgets(s,MAXLINEWIDTH,fp);
  nsamples=0; i=0; 
  while ((i<strlen(s))&&((ch=*(s+i))!=0x0a))
    {
      if (ch==0x09)
	nsamples++;
      i++;
    } 
  samples=malloc(sizeof(char)*nsamples*LONGNAMELENGTH);
  for (n=0; n<=(nsamples-1); n++)
    {
      item=getitemval3(s,n+1,&length,0x09);
      strncpy(samples+n*LONGNAMELENGTH,item,length);
      *(samples+n*LONGNAMELENGTH+length)='\0';
      free(item);
    }

  
  // Load gene info and data.
  nvars=0; genestrs=malloc(sizeof(char)*LONGNAMELENGTH);
  data=malloc(sizeof(double)*nsamples);
  while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
    {
      // Load gene info.
      nvars++; genestrs=realloc(genestrs,sizeof(char)*nvars*LONGNAMELENGTH);
      item=getitemval3(s,0,&length,0x09);
      strncpy(genestrs+(nvars-1)*LONGNAMELENGTH,item,length);
      *(genestrs+(nvars-1)*LONGNAMELENGTH+length)='\0';
      free(item);
      
      // Load data.
      data=realloc(data,sizeof(double)*nvars*nsamples);
      for (i=0; i<=(nsamples-1); i++)
	{
	  item=getitemval3(s,i+1,&length,0x09);	  
	  if (strncmp(item,"NaN",3)==0)
	    val=NAval;
	  else
	    val=atof2(item,length);	  
	  *(data+(nvars-1)*nsamples+i)=val;	  
	  free(item);
	}   
    }
  fclose(fp);
 

  // Replace NaN with the mean of the remaining entries of the gene.  
  for (n=0; n<=(nvars-1); n++)
    {
      double tmp[nsamples];
      k=0;
      for (i=0; i<=(nsamples-1); i++)
	{
	  if ((val=*(data+n*nsamples+i))>=0)
	    {
	      k++; tmp[k-1]=val;
	    }
	}
      val=mean(k,tmp);
      for (i=0; i<=(nsamples-1); i++)
	if (*(data+n*nsamples+i)<0)	
	  *(data+n*nsamples+i)=val;
    }
  
  // Debug
  printf("nvars=%d, nsamples=%d.\n",nvars,nsamples);
  
  
  // Calculate correlation coefficients.
  corrs=malloc(sizeof(double)*nvars*nvars);
  for (i=0; i<=(nvars-2); i++)
    {
      for (j=(i+1); j<=(nvars-1); j++)
	{
	  val=corrcoef(nsamples,data+i*nsamples,data+j*nsamples);
	  *(corrs+i*nvars+j)=val;
	  *(corrs+j*nvars+i)=val;
	}
    }
	

  // Load threshold on correlation coefficient.
  threlwb=atof(*(argv+2));

  // Debug
  printf("threlwb=%.4f\n",threlwb);
    
  isvalid=malloc(sizeof(int)*nvars);
  for (i=0; i<=(nvars-1); i++)    
    *(isvalid+i)=1;      
    

  // Sort the correlation coefficient pairs.
  ntriplets=0; triplets=malloc(sizeof(struct triplet));
  for (i=0; i<=(nvars-2); i++)
    {
      for (j=(i+1); j<=(nvars-1); j++)
	{
	  //if ((inlist(nsubset,subset,*(indices+i))==0)&&(inlist(nsubset,subset,*(indices+j))==0)&&(*(corrs+i*nvars+j)>=threlwb))
	  if ((*(isvalid+i)==1)&&(*(isvalid+j)==1)&&(*(corrs+i*nvars+j)>=threlwb))
	    {
	      ntriplets++;
	      triplets=realloc(triplets,sizeof(struct triplet)*ntriplets);
	      (triplets+ntriplets-1)->i=i;
	      (triplets+ntriplets-1)->j=j;
	      (triplets+ntriplets-1)->s=*(corrs+i*nvars+j);
	    }
	}
    }
  qsort(triplets,ntriplets,sizeof(struct triplet),(* triplet_cmp));
  
  // Debug
  printf("ntriplets=%ld\n",ntriplets);
  printf("%d %d %f\n",(triplets+0)->i,(triplets+0)->j,(triplets+0)->s);
  

  val=(double)ntriplets/(double)nintervals; val=floor(val);
  incsize=(int)val;

    

  // Get maximal cliques.
  // Vary the corrthre from the highest value to the lower bound.  Find the cliques from the graph.
  // Exclude the members in the large cluster.

  
  clids=malloc(sizeof(int)*nvars);
  for (i=0; i<=(nvars-1); i++)
    *(clids+i)=-1;
  newclids=malloc(sizeof(int)*nvars);
  ncliques=0; flag=1; curind=0; corrthre=(triplets+curind)->s;
  pairs=malloc(sizeof(struct pair)*nvars);

  while (flag==1)
    {      
      int nnewcliques, *newconn, *newids, *template, *nconns, lwdn, *tmpnconns;
      
      // Try to merge the cliques obtained from previous steps.
      if (ncliques>0)
	{
	  int *conn, *clcmps;

	  // Debug
	  //printf("merge existing clusters\n");
      
	  conn=malloc(sizeof(int)*ncliques*ncliques);
	  for (i=0; i<ncliques; i++)
	    for (j=0; j<ncliques; j++)
	      *(conn+i*ncliques+j)=0;
	  for (i=0; i<ncliques; i++)
	    *(conn+i*ncliques+i)=1;
	  for (i=0; i<=(ncliques-2); i++)
	    {
	      for (j=(i+1); j<=(ncliques-1); j++)
		{
		  k=0; flag2=1;
		  while ((k<=(nvars-1))&&(flag2==1))
		    {
		      if (*(clids+k)==i)
			{
			  l=0;
			  while ((l<=(nvars-1))&&(flag2==1))
			    {
			      if (*(clids+l)==j)
				{
				  if (*(corrs+k*nvars+l)<corrthre)
				    flag2=0;
				}
			      l++;
			    }
			}
		      k++;
		    }
		  if (flag2==1)
		    {
		      int ff=1, n1=0, n2=0, n3=0, n4=0;

		      if (ff==1)
			{
			  *(conn+i*ncliques+j)=1; *(conn+j*ncliques+i)=1;
			}
		    }
		}
	    }	  
	  	  
	  clcmps=find_undir_conn_comps(ncliques,conn);	  
	

	  for (i=0; i<=(nvars-1); i++)
	    {
	      j=*(clids+i); 
	      if (j>=0)
		{
		  k=*(clcmps+j); *(clids+i)=k;
		}
	    }	 	  

	  ncliques=0;
	  for (i=0; i<=(nvars-1); i++)
	    ncliques=max(ncliques,*(clids+i)+1);
	  free(clcmps);	free(conn);
	
	}

      // Sort the free genes by their connectivity.
      for (i=0; i<=(nvars-1); i++)
	{
	  (pairs+i)->i=i;
	  (pairs+i)->j=0;
	  if ((*(clids+i)<0)&&(*(isvalid+i)==1))
	    for (j=0; j<=(nvars-1); j++)
	      if ((j!=i)&&(*(corrs+i*nvars+j)>=corrthre))
		(pairs+i)->j+=1;	    
	}
      qsort(pairs,nvars,sizeof(struct pair),(* pair_cmp));
      

      // Debug
      //printf("after merging existing cliques, %d %d\n",*(clids+1251),*(clids+1773));


      // Try to assign a newly emerged gene to an existing clique.      
      // If there are multiple candidate cliques then choose the largest one.
      if (ncliques>0)
	{
	  // Debug
 	  //printf("assign newly emerged genes to existing clusters.\n");

	  for (n=0; n<=(nvars-1); n++)
	    {
	      if ((pairs+n)->j>0)
		{
		  int ncandidates=0;	      
		  struct pair *candidates;
		  i=(pairs+n)->i;
		  candidates=malloc(sizeof(struct pair));
		  for (j=0; j<=(ncliques-1); j++)
		    {
		      k=0; flag2=1; l=0;
		      while ((k<=(nvars-1))&&(flag2==1))
			{				        
			  if (*(clids+k)==j)
			    {
			      l++;									
			      if (*(corrs+i*nvars+k)<corrthre)
				flag2=0;
			    }			
			  k++;
			}
		      if (flag2==1)
			{
			  int ff=0;
			  l=0; m=0;
			  for (k=0; k<=(nvars-1); k++)
			    {
			      if (*(clids+k)==j)
				{
				  l++;				  
				}
			    }
			  if (m<1)
			    ff=1;			    
			  if (ff==1)
			    {				
			      ncandidates++;
			      candidates=realloc(candidates,sizeof(struct pair)*ncandidates);
			      (candidates+ncandidates-1)->i=j;
			      (candidates+ncandidates-1)->j=l;
			    }
			}
		    }
		
		  if (ncandidates>0)
		    {
		      int ff;

		      qsort(candidates,ncandidates,sizeof(struct pair),(* pair_cmp));
		      *(clids+i)=(candidates+0)->i;		    
		    }
		  free(candidates);
		}
	    }
	}


      // Try to generate new cliques from the newly emerged genes.
      
      // Debug
      //printf("generate new cliques from newly emerged genes.\n");


      // Sort the newly emerged genes by their connectivity to newly emerged genes.
      for (i=0; i<=(nvars-1); i++)
	{
	  (pairs+i)->i=i;
	  (pairs+i)->j=0;
	  if ((*(clids+i)<0)&&(*(isvalid+i)==1))
	    for (j=0; j<=(nvars-1); j++)
	      if ((j!=i)&&(*(clids+j)<0)&&(*(isvalid+j)==1)&&(*(corrs+i*nvars+j)>=corrthre))
		(pairs+i)->j+=1;	    
	}
      qsort(pairs,nvars,sizeof(struct pair),(* pair_cmp));
      k=0; newids=malloc(sizeof(int)*nvars);
      for (n=0; n<=(nvars-1); n++)
	{
	  i=(pairs+n)->i;
	  if ((pairs+n)->j<=0)
	    *(newclids+i)=-1;
	  else
	    {
	      *(newclids+i)=k; 
	      *(newids+k)=i;
	      k++;
	    }
	}
      flag2=1; nnewcliques=k;


      // Calculate the connection matrix of the cliques.
      // Initially all cliques are singletons.
      newconn=malloc(sizeof(int)*nvars*nvars);
      template=malloc(sizeof(int)*nvars*nvars);
      nconns=malloc(sizeof(int)*nvars);
      tmpnconns=malloc(sizeof(int)*nvars);
      for (i=0; i<=(nvars*nvars-1); i++)
	*(newconn+i)=0;
      for (i=0; i<=(nvars-1); i++)
	{
	  *(newconn+i*nvars+i)=1; *(nconns+i)=0;
	}
      
      for (i=0; i<=(nnewcliques-2); i++)
	{
	  for (j=(i+1); j<=(nnewcliques-1); j++)
	    {
	      int ind1=*(newids+i), ind2=*(newids+j);
	      if (*(corrs+ind1*nvars+ind2)>=corrthre)
		{
		  int ff=1;		  
		  if (ff==1)
		    {
		      *(newconn+i*nvars+j)=1;
		      *(newconn+j*nvars+i)=1;
		      *(nconns+i)+=1; *(nconns+j)+=1;
		    }
		}
	    }
	}
      lwdn=0;
     
      while (flag2==1)
	{
	  int mergeable=0, ind1=-1, ind2=-1;

	  // Find the lower bound where nconns>0.
	  i=0; lwdn=-1;
	  while ((i<=(nnewcliques-1))&&(lwdn<0))
	    {
	      if (*(nconns+i)>0)
		lwdn=i;
	      i++;
	    }

	  // Merge cliques.
	  // Highly connected genes are merged first.
	  // Find two cliques that are mergeable.
	  mergeable=0; i=lwdn;
	  while ((i<=(nnewcliques-2))&&(mergeable==0))
	    {
	      if (*(nconns+i)>0)
		{
		  j=i+1;
		  while ((j<=(nnewcliques-1))&&(mergeable==0))
		    {
		      if ((*(nconns+j)>0)&&(*(newconn+i*nvars+j)==1))
			{
			  ind1=i; ind2=j; mergeable=1;
			}
		      j++;
		    }
		}
	      i++;
	    }

	  // Merge the two cliques ind1 and ind2.
	  if (mergeable==1)
	    {	     	     
	      // Update the connection information.
	      for (i=0; i<=(nvars*nvars-1); i++)
		*(template+i)=0;
	      for (i=0; i<=(nvars-1); i++)
		*(template+i*nvars+i)=1;
	      for (i=0; i<=(nvars-1); i++)		
		*(tmpnconns+i)=0;

	      // Only need to update the connections involved in ind1 and ind2.
	      
	      // Copy the connection matrix of the cliques which are neither ind1 nor ind2.
	      // The cliques with indices > ind2 will be shifted 1 to the left.
	      for (i=0; i<=(nnewcliques-2); i++)
		{
		  for (j=(i+1); j<=(nnewcliques-1); j++)
		    {
		      if ((i!=ind2)&&(i!=ind1)&&(j!=ind2)&&(j!=ind1))
			{
			  int newi=i, newj=j;
			  if (i>ind2)
			    newi=i-1;
			  if (j>ind2)
			    newj=j-1;
			  *(template+newi*nvars+newj)=*(newconn+i*nvars+j);
			  *(template+newj*nvars+newi)=*(newconn+i*nvars+j);
			  
			}
		    }
		}

	      for (i=0; i<=(nnewcliques-1); i++)
		{
		  int newi=i;
		  if (i>ind2)
		    newi=i-1;
		  *(tmpnconns+newi)=*(nconns+i);
		}

	      // Check the connection of existing cliques with the new clique ind1+ind2.
	      for (i=0; i<=(nnewcliques-1); i++)
		{
		  if ((i!=ind1)&&(i!=ind2))
		    {
		      int newi=i;
		      if (i>ind2)
			newi=i-1;		      

		      // If clique i connects to cliques ind1 and ind2, then clique i connects to the new clique ind1+ind2.
		      // The number of connected cliques to clique i is reduced by 1.
		      // The number of connected cliques to clique ind1 is reduced by 1.
		      if ((*(newconn+i*nvars+ind1)==1)&&(*(newconn+i*nvars+ind2)==1))
			{
			  *(template+newi*nvars+ind1)=1;
			  *(template+ind1*nvars+newi)=1;			  
			  *(tmpnconns+newi)-=1;
			  *(tmpnconns+ind1)-=1;			  
			}
		      
		      // Otherwise clique i does not connect to the new clique ind1+ind2.
		      // If clique i connects to ind1 or ind2, then the number of connected cliques to clique i is reduced by 1.
		      // If clique ind1 connects to clique i, then the number of connected cliques to clique ind1 is reduced by 1.
		      else
			{
			  *(template+newi*nvars+ind1)=0;
			  *(template+ind1*nvars+newi)=0;
			  if ((*(newconn+i*nvars+ind1)==1)||(*(newconn+i*nvars+ind2)==1))
			    *(tmpnconns+newi)-=1;
			  if (*(newconn+i*nvars+ind1)==1)
			    *(tmpnconns+ind1)-=1;
			}
		    }
		}			  	      
	      
	      for (i=0; i<=(nvars*nvars-1); i++)
		*(newconn+i)=*(template+i);
	      for (i=0; i<=(nvars-1); i++)
		*(nconns+i)=*(tmpnconns+i);
	      

	      			  
	      // Update the clique indices.
	      for (i=0; i<=(nvars-1); i++)		
		if (*(newclids+i)==ind2)
		  *(newclids+i)=ind1;		
	      for (i=0; i<=(nvars-1); i++)
		if ((k=*(newclids+i))>ind2)
		  *(newclids+i)=k-1;
	      nnewcliques-=1;	      

	    }
	  // No two cliques are mergeable: stop.
	  else
	    flag2=0;
	}

      free(newconn); free(template); free(nconns); free(tmpnconns);
      

      // Add new cliques to the records.
      for (i=0; i<=(nvars-1); i++)
	if (*(newclids+i)>=0)
	  *(clids+i)=*(newclids+i)+ncliques;
      ncliques+=nnewcliques;

      // Sort cliques by their sizes.
      for (i=0; i<=(ncliques-1); i++)
	{
	  (pairs+i)->i=i;
	  (pairs+i)->j=0;
	  for (j=0; j<=(nvars-1); j++)
	    {
	      k=*(clids+j);
	      if (k==i)
		(pairs+i)->j+=1;
	    }
	}
      qsort(pairs,ncliques,sizeof(struct pair),(* pair_cmp));
      for (i=0; i<=(nvars-1); i++)
	{
	  if ((j=*(clids+i))>=0)
	    {
	      k=0; l=-1;
	      while ((k<=(ncliques-1))&&(l<0))
		{
		  if (j==(pairs+k)->i)
		    l=k;
		  k++;
		}
	      *(clids+i)=l;
	    }
	}
	 

      // Debug
      printf("curind=%d, indices=(%d,%d), corrthre=%f, ncliques=%d, ",curind,(triplets+curind)->i,(triplets+curind)->j,corrthre,ncliques);      
      printf("\n");
	       
      

      curind+=incsize;

      if (curind>=ntriplets)
	flag=0;

      else
	{
	  corrthre=(triplets+curind)->s;
	  if (corrthre<threlwb)
	    flag=0;
	}
      

    }

  // Sort cliques by their sizes.
  for (n=0; n<=(ncliques-1); n++)
    {
      (pairs+n)->i=n;
      (pairs+n)->j=0;
      for (i=0; i<=(nvars-1); i++)	
	if (*(clids+i)==n)
	  (pairs+n)->j+=1;
    }
  qsort(pairs,ncliques,sizeof(struct pair),(* pair_cmp));

  
  // Report the cliques.
  if ((fout=fopen(*(argv+3),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+3));
      return -1;
    }
  for (n=0; n<=(ncliques-1); n++)
    {
      if ((pairs+n)->j>1)
	{
	  i=(pairs+n)->i;
	  fprintf(fout,"clique %d has %d genes\n",n+1,(pairs+n)->j);
	  for (j=0; j<=(nvars-1); j++)	    
	    if (*(clids+j)==i)
	      fprintf(fout,"%d %s\n",j+1,genestrs+j*LONGNAMELENGTH);	      
	  fprintf(fout,"\n");
	}
    }
  fclose(fout);


  // Incrementally merge clusters.
  // Start with cliques as clusters.
  // Two clusters are mergeable if the inter-cluster correlation mean is not too small compared to the intra-cluster correlation mean.
  // In each run, choose two largest mergeable clusters to merge.
  // Stop when no clusters are mergeable.

  // Load the threshold of correlation coefficient difference.
  diffthre=atof(*(argv+4));

  // Initialize clusters.
  nclus=ncliques; cluids=malloc(sizeof(int)*nvars);
  for (i=0; i<=(nvars-1); i++)
    {
      *(cluids+i)=*(clids+i); nclus=max(nclus,*(clids+i)+1);
    }  
  
  
  // Merge clusters.
  flag=1;
  while (flag==1)
    {      
      int ind1=-1, ind2=-1, mergeable=0;

      // Sort clusters by sizes.
      for (i=0; i<=(nclus-1); i++)
	{
	  (pairs+i)->i=i; (pairs+i)->j=0;
	}
      for (i=0; i<=(nvars-1); i++)
	{
	  j=*(cluids+i); (pairs+j)->j+=1;
	}
      qsort(pairs,nclus,sizeof(struct pair),(* pair_cmp));
      for (i=0; i<=(nvars-1); i++)
	{
	  j=*(cluids+i); k=0; l=-1;
	  while ((k<=(nclus-1))&&(l<0))
	    {
	      if (j==(pairs+k)->i)
		l=k;
	      k++;
	    }
	  *(cluids+i)=l;
	}

      // Choose two mergeable clusters.
      mergeable=0; i=0;
      while ((i<=(nclus-2))&&(mergeable==0))
	{
	  j=(i+1);
	  while ((j<=(nclus-1))&&(mergeable==0))
	    {
	      int n1=0, n2=0, *subset1, *subset2;
	      double val1=0, val2=0, val3=0;

	      subset1=malloc(sizeof(int)*nvars);
	      subset2=malloc(sizeof(int)*nvars);
	      for (k=0; k<=(nvars-1); k++)
		{
		  if (*(cluids+k)==i)
		    {
		      n1++; *(subset1+n1-1)=k;
		    }
		  else if (*(cluids+k)==j)
		    {
		      n2++; *(subset2+n2-1)=k;
		    }
		}
	      
	      // Calculate the intra-cluster mean correlation coefficients.	      
	      for (k=0; k<=(n1-2); k++)
		{
		  for (l=(k+1); l<=(n1-1); l++)
		    {
		      val=*(corrs+*(subset1+k)*nvars+*(subset1+l));
		      val1+=val;
		    }
		}
	      for (k=0; k<=(n2-2); k++)
		{
		  for (l=(k+1); l<=(n2-1); l++)
		    {
		      val=*(corrs+*(subset2+k)*nvars+*(subset2+l));
		      val1+=val;
		    }
		}
	      val3+=val1;
	      val1/=(double)((n1*(n1-1)/2)+(n2*(n2-1)/2));
	      
      
	      // Calculate the inter-cluster mean correlation coefficients.
	      for (k=0; k<=(n1-1); k++)
		{
		  for (l=0; l<=(n2-1); l++)
		    {
		      val=*(corrs+*(subset1+k)*nvars+*(subset2+l));
		      val2+=val;
		    }
		}
	      val3+=val2;
	      val2/=(double)(n1*n2);	      
	      val3/=(double)((n1*(n1-1)/2)+(n2*(n2-1)/2)+n1*n2);

	      // Check if the difference of the mean correlation coefficients is small.
	      // Don't merge singletons.
	      // Also don't merge if the joint cluster mean correlation <= threlwb.      
	      if (((val=val1-val2)<=diffthre)&&(n1>1)&&(n2>1)&&(val3>=threlwb))
		{
		  int ff=1, m1=0, m2=0, m3=0, m4=0;
		  
		  if (ff==1)
		    {		  
		      mergeable=1; ind1=i; ind2=j;

		      // Debug
		      printf("ind1=%d, ind2=%d, n1=%d, n2=%d, val1=%f, val2=%f, val3=%f, diff=%f\n",ind1,ind2,n1,n2,val1,val2,val3,val);
		    }

		}
	      j++; free(subset1); free(subset2);
	    }
	  i++;
	}
		  

      // Merge two clusters.
      if (mergeable==1)
	{
	  // Assign the members of cluster j to cluster i.
	  // Shift the ids of the clusters > ind1.
	  for (i=0; i<=(nvars-1); i++)
	    {
	      if (*(cluids+i)==ind2)
		*(cluids+i)=ind1;
	      else if (*(cluids+i)>ind2)
		*(cluids+i)-=1;
	    }
	  nclus--;
	}	  	  	  	

      // Otherwise stop.
      else
	flag=0;
    }
	      
  
  // Debug
  printf("ncliques=%d, nclus=%d\n",ncliques,nclus);

  
  // Report the clusters.
  if ((fout=fopen(*(argv+5),"w"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+5));
      return -1;
    }

  // Sort clusters by sizes.
  for (i=0; i<=(nclus-1); i++)
    {
      (pairs+i)->i=i; (pairs+i)->j=0;
    }
  for (i=0; i<=(nvars-1); i++)
    {
      j=*(cluids+i); (pairs+j)->j+=1;
    }
  qsort(pairs,nclus,sizeof(struct pair),(* pair_cmp));
  for (i=0; i<=(nvars-1); i++)
    {
      j=*(cluids+i); k=0; l=-1;
      while ((k<=(nclus-1))&&(l<0))
	{
	  if (j==(pairs+k)->i)
	    l=k;
	  k++;
	}
      *(cluids+i)=l;
    }


  for (n=0; n<=(nclus-1); n++)
    {
      if ((pairs+n)->j>1)
	{
	  i=n;
	  fprintf(fout,"cluster %d has %d genes\n",n+1,(pairs+n)->j);
	  for (j=0; j<=(nvars-1); j++)	    
	    if (*(cluids+j)==i)		
	      fprintf(fout,"%d %s\n",j+1,genestrs+j*LONGNAMELENGTH);	      
	  fprintf(fout,"\n");
	}
    }
  fclose(fout);


	  
  return 0;
}


