% Partition the CNV array data into segments.
% The CNV array data is too large to directly partition.  Do it in two steps.
% Step 1: Generate a submatrix by sampling every 10 probes in the sorted matrix.  Perform recursive partitioning on the lower resolution matrix.
% Step 2: From the partitioning results of the lower resolution matrix, refine the partition boundaries on the higher resolution matrix.
% Assume the probes of the input data are sorted by chromosomal coordinates.
% Difference from partition_CNV.m: (1)Generate a partition for each sample separately.  (2)Incur probabilistic quantization to get the fractional counts of tristates (-1: deletion, 0: no change, 1: amplification), instead of doing deterministic quantization with a fixed threshold.
% Difference from partition_CNV2.m: (1)Apply a nonlinear adjustment to rescale cdf values such that the total probabilities of amplification and deletion matches the observations from the cnv probe data, (2)Explicitly search recurrent amplifications or deletions on small fragments.  These small fragments may not be isolated from the naive Bayes algorithm.

function [segmentinfo, smallfrags] = partition_CNV3(nprobes, nsamples, probes, samples, cdfdata, lowressize, numthre, valthre, ga, signalthre, recurthre, distthre)


% Get the chromosome for each probe.
clear chrprobes;
for i=1:nprobes
chrprobes(i)=probes{i}.chr;
end


% Debug
fprintf('low resolution.\n');


% Apply recursive partition on low resolution data for each chromosome.
clear lowrespartitions;
for n=1:24

% Find the probes in chromosome n.
subset=find(chrprobes==n);
subcdfdata=cdfdata(subset,:);

% If the probe set size <= lowressize, then directly use the high resolution data.
if (length(subset)<=lowressize)
nint=1;
lowresset=1:length(subset);
lowrescdfdata=subcdfdata;
% Otherwise sample the probes to make the data size ~ lowressize.
else
nint=ceil(length(subset)/lowressize);
low=1; upp=floor(length(subset)/nint); upp=(upp-1)*nint+1;
lowresset=low:nint:upp;
lowrescdfdata=subcdfdata(lowresset,:);
end


% Apply probabilistic quantization to get the fractional counts of tristates for each probe on each sample.
% For NA, the fractional counts are (0,0,0).
% qdata(:,:,1): -1.  qdata(:,:,2): 0.  qdata(:,:,3): 1.
% Apply a nonlinear adjustment to rescale the cdf values.
% Let x be a cdf value.  P(s=+1|x,ga)=x^ga/(1-log(x^ga)).  P(s=-1|x,ga)=(1-x)^ga/(1-log((1-x)^ga)).

qdata=zeros(length(lowrescdfdata(:,1)),length(lowrescdfdata(1,:)),3);
set1=find(~((lowrescdfdata>=0)|(lowrescdfdata<0)));
set2=find(lowrescdfdata<=0);
set3=find(lowrescdfdata>=1);
set4=find((lowrescdfdata>0)&(lowrescdfdata<1));
tmpmat=zeros(size(lowrescdfdata));
tmpmat(set2)=1; qdata(:,:,1)=qdata(:,:,1)+tmpmat;
tmpmat=zeros(size(lowrescdfdata));
tmpmat(set3)=1; qdata(:,:,3)=qdata(:,:,3)+tmpmat;
tmpmat=zeros(size(lowrescdfdata));
tmpmat(set4)=((1-lowrescdfdata(set4)).^ga)./(1-log((1-lowrescdfdata(set4)).^ga));
qdata(:,:,1)=qdata(:,:,1)+tmpmat;
tmpmat=zeros(size(lowrescdfdata));
tmpmat(set4)=((lowrescdfdata(set4)).^ga)./(1-log((lowrescdfdata(set4)).^ga));
qdata(:,:,3)=qdata(:,:,3)+tmpmat;
set0=union(set2,union(set3,set4));
q1=qdata(:,:,1); q2=zeros(size(lowrescdfdata)); q3=qdata(:,:,3); 
q2(set0)=1-q1(set0)-q3(set0);
qdata(:,:,2)=qdata(:,:,2)+q2;


% Incur the recursive partitioning algorithm.
% Do it for each sample separately.

[nsubprobes,nsubsamples]=size(lowrescdfdata);
for sampleind=1:nsamples
startind=1; endind=nsubprobes;
labels=zeros(1,nsubprobes);
clear fraccnts;
for i=1:nsubprobes
fraccnts(i,:)=[qdata(i,sampleind,1) qdata(i,sampleind,2) qdata(i,sampleind,3)];
end
labels=recurse_partition3(fraccnts,labels,startind,endind,numthre,valthre);

% Find the partitioning boundaries.
clear bds;
bds=[startind endind];
for k=0:max(labels)
if (length(labels==k)>0)
l=max(find(labels==k));
bds=union(bds,l);
end
end
bds=sort(bds);


% Convert the partitioning boundaries in the low resolution probes into the high resolution probes.
% Each boundary probe in the low resolution set corresponds to a list of probes in the high resolution set.
lowrespartitions{n}{sampleind}.nbds=length(bds); lowrespartitions{n}{sampleind}.bds={};
for i=1:length(bds)
j=bds(i); 

% Subsampling does occur.
ss=((j-1)*nint+1):((j-1)*nint+nint);
lowrespartitions{n}{sampleind}.bds{i}=ss;
end

% end of for sampleind=1:nsamples
end

% end of for n=1:24
end


% Debug
fprintf('high resolution.\n');


% Refine the partition boundaries from low resolution data.
clear highrespartitions;
for n=1:24

subset=find(chrprobes==n);
subcdfdata=cdfdata(subset,:);
qdata=zeros(length(subcdfdata(:,1)),length(subcdfdata(1,:)),3);
set1=find(~((subcdfdata>=0)|(subcdfdata<0)));
set2=find(subcdfdata<=0);
set3=find(subcdfdata>=1);
set4=find((subcdfdata>0)&(subcdfdata<1));
tmpmat=zeros(size(subcdfdata));
tmpmat(set2)=1; qdata(:,:,1)=qdata(:,:,1)+tmpmat;
tmpmat=zeros(size(subcdfdata));
tmpmat(set3)=1; qdata(:,:,3)=qdata(:,:,3)+tmpmat;
tmpmat=zeros(size(subcdfdata));
tmpmat(set4)=((1-subcdfdata(set4)).^ga)./(1-log((1-subcdfdata(set4)).^ga));
qdata(:,:,1)=qdata(:,:,1)+tmpmat;
tmpmat=zeros(size(subcdfdata));
tmpmat(set4)=((subcdfdata(set4)).^ga)./(1-log((subcdfdata(set4)).^ga));
qdata(:,:,3)=qdata(:,:,3)+tmpmat;
set0=union(set2,union(set3,set4));
q1=qdata(:,:,1); q2=zeros(size(subcdfdata)); q3=qdata(:,:,3); 
q2(set0)=1-q1(set0)-q3(set0);
qdata(:,:,2)=qdata(:,:,2)+q2;


for sampleind=1:nsamples

highrespartitions{n}{sampleind}.nbds=lowrespartitions{n}{sampleind}.nbds;

for i=1:lowrespartitions{n}{sampleind}.nbds

% First boundary: the first probe.
if (i==1)
highrespartitions{n}{sampleind}.bds(i)=1;

% Last boundary: the last probe.
elseif (i==lowrespartitions{n}{sampleind}.nbds)
highrespartitions{n}{sampleind}.bds(i)=length(subset);

% Other boundaries: shift the boundary in the range provided by lowrespartitions, find the one maximizing the likelihood score.
else
low=min(lowrespartitions{n}{sampleind}.bds{i-1}); upp=max(lowrespartitions{n}{sampleind}.bds{i+1});

ls=-inf*ones(1,length(subset));
for j=1:length(lowrespartitions{n}{sampleind}.bds{i})
k=lowrespartitions{n}{sampleind}.bds{i}(j);
clear fraccnts1 fraccnts2;
for l=low:k
fraccnts1(l-low+1,:)=[qdata(l,sampleind,1) qdata(l,sampleind,2) qdata(l,sampleind,3)];
end
l1=naive_Bayes_loglikelihood2(fraccnts1);
for l=(k+1):upp
fraccnts2(l-k,:)=[qdata(l,sampleind,1) qdata(l,sampleind,2) qdata(l,sampleind,3)];
end
l2=naive_Bayes_loglikelihood2(fraccnts2);
ls(k)=l1+l2;
end

k=find(ls>=max(ls)); k=k(1);
highrespartitions{n}{sampleind}.bds(i)=k;

% end of if (i==1).
end

% end of for i=1:lowrespartitions{n}{sampleind}.nbds.
end

% end of for sampleind=1:nsamples
end

% end of for n=1:24.
end

% Debug
fprintf('generate segments.\n');


% Find small fragments that are recurrently amplified or deleted across many samples.
% These small fragments may not be picked up by the naive Bayes algorith.  For example, the fragment containing EGFR on chr7 on GBM data.
% Find the probes that have high or low values in more than samplefracthre of samples.
% Combine adjacent probes with similar profiles of amplifications and deletions.
% Stop when all candidate probes are covered.
clear smallfrags;
for n=1:24

subset=find(chrprobes==n);
subcdfdata=cdfdata(subset,:);
qdata=zeros(length(subcdfdata(:,1)),length(subcdfdata(1,:)),3);
set1=find(~((subcdfdata>=0)|(subcdfdata<0)));
set2=find(subcdfdata<=0);
set3=find(subcdfdata>=1);
set4=find((subcdfdata>0)&(subcdfdata<1));
tmpmat=zeros(size(subcdfdata));
tmpmat(set2)=1; qdata(:,:,1)=qdata(:,:,1)+tmpmat;
tmpmat=zeros(size(subcdfdata));
tmpmat(set3)=1; qdata(:,:,3)=qdata(:,:,3)+tmpmat;
tmpmat=zeros(size(subcdfdata));
tmpmat(set4)=((1-subcdfdata(set4)).^ga)./(1-log((1-subcdfdata(set4)).^ga));
qdata(:,:,1)=qdata(:,:,1)+tmpmat;
tmpmat=zeros(size(subcdfdata));
tmpmat(set4)=((subcdfdata(set4)).^ga)./(1-log((subcdfdata(set4)).^ga));
qdata(:,:,3)=qdata(:,:,3)+tmpmat;
set0=union(set2,union(set3,set4));
q1=qdata(:,:,1); q2=zeros(size(subcdfdata)); q3=qdata(:,:,3); 
q2(set0)=1-q1(set0)-q3(set0);
qdata(:,:,2)=qdata(:,:,2)+q2;

% Find the probes that are amplified or deleted.
cnvchange=zeros(size(subcdfdata));
set1=find(qdata(:,:,3)>=signalthre);
cnvchange(set1)=1;
set1=find(qdata(:,:,1)>=signalthre);
cnvchange(set1)=-1;


% Find the probes that are recurrently amplified or deleted.
% The recurrency frequency depends on the background frequency of amplifications and deletions on a chromosome.
% For instance, chr7 has a high amplification frequency.
% Calculate the binomial probability that a probe has >= k samples amplified or deleted.
p1=sum(sum(qdata(:,:,3)>=signalthre))/(length(subcdfdata(:,1))*length(subcdfdata(1,:)));
p2=sum(sum(qdata(:,:,1)>=signalthre))/(length(subcdfdata(:,1))*length(subcdfdata(1,:)));

set1=find(sum(transpose(cnvchange==1))>=(nsamples*p1*2));
vals1=zeros(1,length(set1));
for i=1:length(set1)
j=set1(i); k=sum(cnvchange(j,:)==1);
for l=k:nsamples
val2=logchoose(nsamples,l)+l*log(p1)+(nsamples-l)*log(1-p1);
vals1(i)=vals1(i)+exp(val2);
end
end
tmp=find(vals1<=recurthre); 
if (length(tmp)>0)
set1=set1(tmp); vals1=vals1(tmp);
else
set1=[]; vals1=[];
end

set2=find(sum(transpose(cnvchange==-1))>=(nsamples*p2*2));
vals2=zeros(1,length(set2));
for i=1:length(set2)
j=set2(i); k=sum(cnvchange(j,:)==-1);
for l=k:nsamples
val2=logchoose(nsamples,l)+l*log(p2)+(nsamples-l)*log(1-p2);
vals2(i)=vals2(i)+exp(val2);
end
end
tmp=find(vals2<=recurthre); 
if (length(tmp)>0)
set2=set2(tmp); vals2=vals2(tmp);
else
set2=[]; vals2=[];
end


% Use the probes with small p-values as seeds to generate currently amplified/deleted fragments.
% Combine seed probes with short distances.
jset=union(set1,set2); chain=zeros(1,length(jset));
for i=1:(length(jset)-1)
j1=jset(i); j2=jset(i+1);
if ((j2-j1)<=distthre)
chain(i)=1;
end
end

smallfrags{n}.nfrags=0;
smallfrags{n}.frags={};
i=1;
while (i<=length(jset))
% A new fragment.
if ((i==1)|((i>1)&(chain(i-1)==0)))
smallfrags{n}.nfrags=smallfrags{n}.nfrags+1;
smallfrags{n}.frags{smallfrags{n}.nfrags}=jset(i);
% An old fragment.
else
smallfrags{n}.frags{smallfrags{n}.nfrags}=union(smallfrags{n}.frags{smallfrags{n}.nfrags},jset(i));
end
i=i+1;
end

for i=1:smallfrags{n}.nfrags
for j=1:length(smallfrags{n}.frags{i})
k=smallfrags{n}.frags{i}(j);
smallfrags{n}.frags{i}(j)=subset(k);
end
end

% end of for n=1:24.
end


% Small fragments are filterd by combining two CNV data.



% Generate segments from the partitions.
% Notice each sample incurs a distinct partition.
clear segmentinfo;
for sampleind=1:nsamples
nsegments=0; clear segments;
for n=1:24
subset=find(chrprobes==n);
for i=1:(highrespartitions{n}{sampleind}.nbds-1)
nsegments=nsegments+1;
segments{nsegments}.chr=n;
j=highrespartitions{n}{sampleind}.bds(i); j=subset(j);
if (i>1)
j=j+1;
end
k=highrespartitions{n}{sampleind}.bds(i+1); k=subset(k);
segments{nsegments}.startcoord=probes{j}.start;
segments{nsegments}.endcoord=probes{k}.end;
segments{nsegments}.startprobe=j;
segments{nsegments}.endprobe=k;
end
end
segmentinfo{sampleind}.nsegments=nsegments;
segmentinfo{sampleind}.segments=segments;
end
